# coding:utf-8

from http.server import ThreadingHTTPServer
from threading import Thread
from time import sleep
import unittest
from unittest import mock

from requests.exceptions import ConnectionError

from xserver.http import proxy


class FakeProxy():
    def __init__(self, host: str, port: int):
        self.target_url = "https://example.com/"
        self.listen_address = (host, port)

    def run(self):
        self.httpd = ThreadingHTTPServer(
            self.listen_address, lambda *args: proxy.HttpProxy(
                *args, create_request_proxy=proxy.RequestProxy.create,
                target_url=self.target_url))
        self.httpd.serve_forever()


class TestResponseProxy(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_generator(self):
        response = proxy.ResponseProxy.make_ok_response(b"test")
        self.assertEqual(response.status_code, 200)
        for chunk in response.generator:
            self.assertEqual(chunk, b"test")

    def test_close(self):
        self.assertIsNone(proxy.ResponseProxy(status_code=304, headers=proxy.HeaderSequence()).close())  # noqa:E501

    def test_set_cookie(self):
        self.assertIsNone(proxy.ResponseProxy(status_code=404, headers=proxy.HeaderSequence()).set_cookie("test", "unit"))  # noqa:E501

    def test_redirect(self):
        response = proxy.ResponseProxy.redirect()
        self.assertEqual(response.status_code, 302)
        for header in response.headers:
            self.assertEqual(header, (proxy.Headers.LOCATION.http2, "/"))


class TestRequestProxyResponse(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        self.fake_response = mock.MagicMock()
        self.fake_response.status_code = 404
        self.fake_response.headers = {"Content-Length": "0"}
        self.fake_response.iter_content.side_effect = [["test"]]
        self.response = proxy.RequestProxyResponse(self.fake_response)

    def tearDown(self):
        pass

    def test_status_code(self):
        self.assertEqual(self.response.status_code, 404)

    def test_headers(self):
        self.assertEqual(len(self.response.headers), 0)

    def test_generator(self):
        for chunk in self.response.generator:
            self.assertEqual(chunk, "test")

    def test_close(self):
        self.assertIsNone(self.response.close())


class TestRequestProxy(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with mock.patch.object(proxy, "Session") as mock_session:
            cls.fake_session = mock.MagicMock()
            mock_session.side_effect = [cls.fake_session]
            cls.request_proxy: proxy.RequestProxy = proxy.RequestProxy("https://example.com/")  # noqa:E501

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_request_GET(self):
        self.assertIsInstance(self.request_proxy.request("test", "GET"), proxy.ResponseProxy)  # noqa:E501

    def test_request_POST(self):
        self.assertIsInstance(self.request_proxy.request("test", "POST"), proxy.ResponseProxy)  # noqa:E501

    def test_request_MethodNotAllowed(self):
        self.assertRaises(proxy.MethodNotAllowed, self.request_proxy.request, "test", "PUT")  # noqa:E501


class TestHttpProxy(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.port = 12345
        cls.host = "localhost"
        cls.url = f"http://{cls.host}:{cls.port}/test"
        cls.fake_proxy = FakeProxy(cls.host, cls.port)
        cls.thread = Thread(target=cls.fake_proxy.run)
        cls.thread.start()
        fake_response = mock.MagicMock()
        fake_response.status_code = 200
        fake_response.headers = {
            "Accept-Ranges": "bytes",
            "Content-Type": "text/html",
            "ETag": '"84238dfc8092e5d9c0dac8ef93371a07:1736799080.121134"',
            "Last-Modified": "Mon, 13 Jan 2025 20:11:20 GMT",
            "Vary": "Accept-Encoding",
            "Cache-Control": "max-age=771",
            "Date": "Wed, 02 Apr 2025 15:49:17 GMT",
            "Alt-Svc": 'h3=":443"; ma=93600,h3-29=":443"; ma=93600,quic=":443"; ma=93600; v="43"',  # noqa:E501
        }
        fake_response.generator = [b'<!doctype html>\n<html>\n<head>\n    <title>Example Domain</title>\n\n    <meta charset="utf-8" />\n    <meta http-equiv="Content-type" content="text/html; charset=utf-8" />\n    <meta name="viewport" content="width=device-width, initial-scale=1" />\n    <style type="text/css">\n    body {\n        background-color: #f0f0f2;\n        margin: 0;\n        padding: 0;\n        font-family: -apple-system, system-ui, BlinkMacSystemFont, "Segoe UI", "Open Sans", "Helvetica Neue", Helvetica, Arial, sans-serif;\n        \n    }\n    div {\n        width: 600px;\n        margin: 5em auto;\n        padding: 2em;\n        background-color: #fdfdff;\n        border-radius: 0.5em;\n        box-shadow: 2px 3px 7px 2px rgba(0,0,0,0.02);\n    }\n    a:link, a:visited {\n        color: #38488f;\n        text-decoration: none;\n    }\n    @media (max-width: 700px) {\n        div {\n            margin: 0 auto;\n            width: auto;\n        }\n    }\n    </style>    \n</head>\n\n<body>\n<div>\n    <h1>Example Domain</h1>\n    <p>This domain is for use in illustrative examples in documents. You may use this\n    domain in literature without prior coordination or asking for permission.</p>\n    <p><a href="https://www.iana.org/domains/example">More information...</a></p>\n</div>\n</body>\n</html>\n']  # noqa:E501
        cls.fake_response = fake_response
        sleep(0.5)

    @classmethod
    def tearDownClass(cls):
        cls.fake_proxy.httpd.shutdown()

    def setUp(self):
        pass

    def tearDown(self):
        pass

    @mock.patch.object(proxy.RequestProxy, "request")
    @mock.patch.object(proxy.HttpProxy, "send_response")
    def test_forward_BrokenPipeError(self, mock_send_response, mock_request):
        mock_send_response.side_effect = [BrokenPipeError()]
        mock_request.side_effect = [self.fake_response]
        self.assertRaises(ConnectionError, proxy.Session().get, self.url)

    @mock.patch.object(proxy.RequestProxy, "request")
    def test_do_GET(self, mock_request):
        mock_request.side_effect = [self.fake_response]
        proxy.Session().get(self.url)

    @mock.patch.object(proxy.RequestProxy, "request")
    def test_do_POST(self, mock_request):
        mock_request.side_effect = [self.fake_response]
        proxy.Session().post(self.url)


if __name__ == "__main__":
    unittest.main()



# coding:utf-8

from socket import SOL_SOCKET
from socket import SO_RCVBUF  # noqa:H306
from socket import SO_SNDBUF
from socket import create_connection  # noqa:H306
from socket import socket
from socket import timeout
from threading import Thread
from typing import Tuple

from xkits_lib.unit import TimeUnit


class ResponseProxy():
    """Socket Response Proxy"""

    def __init__(self, client: socket, server: socket, chunk: int):
        self.__thread: Thread = Thread(target=self.handler)
        self.__client: socket = client
        self.__server: socket = server
        self.__running: bool = False
        self.__chunk: int = chunk

    @property
    def client(self) -> socket:
        return self.__client

    @property
    def server(self) -> socket:
        return self.__server

    @property
    def running(self) -> bool:
        return self.__running

    @property
    def chunk(self) -> int:
        return self.__chunk

    def handler(self):
        try:
            while self.running:
                try:
                    data: bytes = self.server.recv(self.chunk)
                except timeout:
                    continue

                if len(data) == 0:
                    break

                self.client.sendall(data)
        except Exception:
            pass
        finally:
            if self.server.fileno() >= 0:
                self.server.close()
            if self.client.fileno() >= 0:
                self.client.close()

    def start(self):
        self.__running = True
        self.__thread.start()

    def stop(self):
        self.__running = False
        self.__thread.join()


class SockProxy():
    def __init__(self, host: str, port: int, timeout: TimeUnit, chunk: int = 65536):  # noqa:E501
        self.__target: Tuple[str, int] = (host, port)
        self.__timeout: TimeUnit = timeout
        self.__chunk: int = chunk

    @property
    def target(self) -> Tuple[str, int]:
        return self.__target

    @property
    def timeout(self) -> TimeUnit:
        return self.__timeout

    @property
    def chunk(self) -> int:
        return self.__chunk

    def new_connection(self, client: socket, data: bytes):
        server: socket = create_connection(address=self.target)
        server.setsockopt(SOL_SOCKET, SO_RCVBUF, self.chunk)
        server.setsockopt(SOL_SOCKET, SO_SNDBUF, self.chunk)
        client.setsockopt(SOL_SOCKET, SO_RCVBUF, self.chunk)
        client.setsockopt(SOL_SOCKET, SO_SNDBUF, self.chunk)
        server.settimeout(self.timeout)
        client.settimeout(self.timeout)

        response: ResponseProxy = ResponseProxy(client, server, chunk=self.chunk)  # noqa:E501

        try:
            response.start()
            while True:
                if len(data) == 0:
                    break

                server.sendall(data)

                try:
                    data = client.recv(self.chunk)
                except timeout:
                    pass
        except OSError:
            pass
        except Exception:
            pass
        finally:
            response.stop()



# coding:utf-8

import unittest

from xserver.sock import header


class TestRequestHeader(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_parse(self):
        request = header.RequestHeader.parse(b"GET / HTTP/1.1\r\nHost: 127.0.0.1\r\n\r\n")  # noqa:501
        self.assertIsInstance(request, header.RequestHeader)
        assert isinstance(request, header.RequestHeader)

        self.assertEqual(request.request_line.protocol, "HTTP/1.1")
        self.assertEqual(request.request_line.method, "GET")
        self.assertEqual(request.request_line.target, "/")

        self.assertIsInstance(request.headers, header.HeaderMapping)

        self.assertEqual(request.length, 35)


class TestResponseHeader(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_parse(self):
        response = header.ResponseHeader.parse(b"HTTP/1.1 200 OK\r\nLocation: http://example.com/users/123\r\n\r\n")  # noqa:501
        self.assertIsInstance(response, header.ResponseHeader)
        assert isinstance(response, header.ResponseHeader)

        self.assertEqual(response.status_line.protocol, "HTTP/1.1")
        self.assertEqual(response.status_line.status_code, 200)
        self.assertEqual(response.status_line.status_text, "OK")

        self.assertIsInstance(response.headers, header.HeaderMapping)

        self.assertEqual(response.length, 59)


if __name__ == "__main__":
    unittest.main()



# coding:utf-8

from http.server import BaseHTTPRequestHandler
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import MutableMapping
from typing import Optional
from urllib.parse import urljoin

from requests import Response
from requests import Session
from xhtml.header.headers import HeaderSequence
from xhtml.header.headers import Headers


class ProxyError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class MethodNotAllowed(ProxyError):
    def __init__(self) -> None:
        super().__init__("Method Not Allowed")


class ResponseProxy():
    """API Response Proxy"""
    CHUNK_SIZE: int = 1048576  # 1MB

    def __init__(self, status_code: int, headers: HeaderSequence, datas: bytes = b"") -> None:  # noqa:E501
        self.__status_code: int = status_code
        self.__headers: HeaderSequence = headers
        self.__datas: bytes = datas

    @property
    def status_code(self) -> int:
        return self.__status_code

    @property
    def headers(self) -> HeaderSequence:
        return self.__headers

    @property
    def generator(self) -> Generator[bytes, Any, None]:
        yield self.__datas

    def close(self):
        pass

    def set_cookie(self, keyword: str, value: str):
        self.headers.add(Headers.SET_COOKIE.value, f"{keyword}={value}")

    @classmethod
    def make_ok_response(cls, datas: bytes) -> "ResponseProxy":
        headers: HeaderSequence = HeaderSequence([(Headers.CONTENT_LENGTH.value, str(len(datas)))])  # noqa:E501
        return ResponseProxy(status_code=200, headers=headers, datas=datas)

    @classmethod
    def redirect(cls, status_code: int = 302, location: str = "/") -> "ResponseProxy":  # noqa:E501
        headers: HeaderSequence = HeaderSequence([(Headers.LOCATION.value, location)])  # noqa:E501
        return ResponseProxy(status_code=status_code, headers=headers)


class RequestProxyResponse(ResponseProxy):
    """API Request Proxy Response"""

    EXCLUDED_HEADERS = [
        Headers.CONNECTION.http2,
        Headers.CONTENT_ENCODING.http2,
        Headers.CONTENT_LENGTH.http2,
        Headers.TRANSFER_ENCODING.http2,
    ]

    def __init__(self, response: Response) -> None:
        headers: HeaderSequence = HeaderSequence([(k, v) for k, v in response.headers.items() if k.lower() not in self.EXCLUDED_HEADERS])  # noqa:E501
        super().__init__(status_code=response.status_code, headers=headers)
        self.__response: Response = response

    @property
    def generator(self):
        for chunk in self.__response.iter_content(chunk_size=self.CHUNK_SIZE):
            yield chunk

    def close(self):
        self.__response.close()


class RequestProxy():
    """API Request Proxy"""

    EXCLUDED_HEADERS = [
        Headers.CONNECTION.http2,
        Headers.CONTENT_LENGTH.http2,
        Headers.HOST.http2,
        Headers.KEEP_ALIVE.http2,
        Headers.PROXY_AUTHORIZATION.http2,
        Headers.TRANSFER_ENCODING.http2,
        Headers.VIA.http2,
    ]

    def __init__(self, target_url: str) -> None:
        self.__target_url: str = target_url
        self.__session: Session = Session()

    @property
    def target_url(self) -> str:
        return self.__target_url

    @property
    def session(self) -> Session:
        return self.__session

    def urljoin(self, path: str) -> str:
        return urljoin(base=self.target_url, url=path)

    @classmethod
    def filter_headers(cls, headers: MutableMapping[str, str]) -> Dict[str, str]:  # noqa:E501
        return {k: v for k, v in headers.items() if k.lower() not in cls.EXCLUDED_HEADERS}  # noqa:E501

    def request(self, path: str, method: str, data: Optional[bytes] = None,
                headers: Optional[MutableMapping[str, str]] = None
                ) -> RequestProxyResponse:
        url: str = self.urljoin(path.lstrip("/"))
        if method == "GET":
            response = self.session.get(
                url=url,
                data=data,
                headers=headers,
                stream=True
            )
            return RequestProxyResponse(response)
        if method == "POST":
            response = self.session.post(
                url=url,
                data=data,
                headers=headers,
                stream=True
            )
            return RequestProxyResponse(response)
        raise MethodNotAllowed()

    @classmethod
    def create(cls, *args, **kwargs) -> "RequestProxy":
        return cls(target_url=kwargs["target_url"])


class HttpProxy(BaseHTTPRequestHandler):
    def __init__(self, *args, create_request_proxy: Callable[..., RequestProxy], **kwargs):  # noqa:E501
        self.__request_proxy: RequestProxy = create_request_proxy(*args, **kwargs)  # noqa:E501
        super().__init__(*args)

    @property
    def request_proxy(self) -> RequestProxy:
        return self.__request_proxy

    def get_request_data(self):
        content_length = int(self.headers.get("Content-Length", 0))
        return self.rfile.read(content_length) if content_length > 0 else None

    def forward(self, rp: ResponseProxy):
        try:
            self.send_response(rp.status_code)
            for header in rp.headers:
                k: str = header[0]
                v: str = header[1]
                self.send_header(k, v)
            self.end_headers()
            for chunk in rp.generator:
                self.wfile.write(chunk)
                self.wfile.flush()
        except BrokenPipeError:
            pass
        finally:
            rp.close()

    def do_GET(self):
        headers = self.request_proxy.filter_headers(
            {k: v for k, v in self.headers.items()})
        response = self.request_proxy.request(
            path=self.path,
            method="GET",
            data=self.get_request_data(),
            headers=headers)
        self.forward(response)

    def do_POST(self):
        headers = self.request_proxy.filter_headers(
            {k: v for k, v in self.headers.items()})
        response = self.request_proxy.request(
            path=self.path,
            method="POST",
            data=self.get_request_data(),
            headers=headers)
        self.forward(response)



# coding:utf-8

from typing import Iterable
from typing import List
from typing import Optional

from xhtml.header.headers import RequestLine
from xhtml.header.headers import StatusLine
from xhtml.header.headers import HeaderMapping


class RequestHeader():
    def __init__(self, request_line: str, request_headers: Iterable[str], header_length: int):  # noqa:E501
        self.__request_line: RequestLine = RequestLine(request_line)
        self.__headers: HeaderMapping = HeaderMapping.parse(request_headers)
        self.__length: int = header_length

    @property
    def request_line(self) -> RequestLine:
        return self.__request_line

    @property
    def headers(self) -> HeaderMapping:
        return self.__headers

    @property
    def length(self) -> int:
        return self.__length

    @classmethod
    def parse(cls, data: bytes) -> Optional["RequestHeader"]:
        offset: int = data.find(b"\r\n\r\n")
        if offset > 0:
            content: str = data[:offset].decode("utf-8")
            headers: List[str] = content.split("\r\n")
            return cls(headers[0], headers[1:], offset + 4)


class ResponseHeader():
    def __init__(self, status_line: str, response_headers: Iterable[str], header_length: int):  # noqa:E501
        self.__status_line: StatusLine = StatusLine(status_line)
        self.__headers: HeaderMapping = HeaderMapping.parse(response_headers)
        self.__length: int = header_length

    @property
    def status_line(self) -> StatusLine:
        return self.__status_line

    @property
    def headers(self) -> HeaderMapping:
        return self.__headers

    @property
    def length(self) -> int:
        return self.__length

    @classmethod
    def parse(cls, data: bytes) -> Optional["ResponseHeader"]:
        offset: int = data.find(b"\r\n\r\n")
        if offset > 0:
            content: str = data[:offset].decode("utf-8")
            headers: List[str] = content.split("\r\n")
            return cls(headers[0], headers[1:], offset + 4)



# coding:utf-8

import unittest
from unittest import mock

from xserver.sock import proxy


class TestResponseProxy(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        self.fake_client = mock.MagicMock()
        self.fake_server = mock.MagicMock()
        self.fake_thread = mock.MagicMock()
        with mock.patch.object(proxy, "socket") as mock_socket:
            mock_socket.side_effect = [self.fake_client, self.fake_server]
            with mock.patch.object(proxy, "Thread") as mock_thread:
                client = proxy.socket()
                server = proxy.socket()
                self.assertIs(client, self.fake_client)
                self.assertIs(server, self.fake_server)
                mock_thread.side_effect = [self.fake_thread]
                self.proxy = proxy.ResponseProxy(client, server, 65536)

    def tearDown(self):
        pass

    def test_handler(self):
        self.fake_client.fileno.side_effect = [1]
        self.fake_server.fileno.side_effect = [2]
        self.fake_server.recv.side_effect = [proxy.timeout(), b""]
        self.assertIsNone(self.proxy.start())
        self.assertIsNone(self.proxy.handler())
        self.assertIsNone(self.proxy.stop())

    def test_handler_Exception(self):
        self.fake_client.fileno.side_effect = [1]
        self.fake_server.fileno.side_effect = [2]
        self.fake_client.sendall.side_effect = [Exception()]
        self.fake_server.recv.side_effect = [proxy.timeout(), b"test"]
        self.assertIsNone(self.proxy.start())
        self.assertIsNone(self.proxy.handler())
        self.assertIsNone(self.proxy.stop())


class TestSockProxy(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.host = "0.0.0.0"
        cls.port = 12345
        cls.timeout = 60

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        self.proxy = proxy.SockProxy(self.host, self.port, self.timeout)

    def tearDown(self):
        pass

    @mock.patch.object(proxy, "socket")
    @mock.patch.object(proxy, "create_connection")
    @mock.patch.object(proxy.ResponseProxy, "stop", mock.MagicMock())
    @mock.patch.object(proxy.ResponseProxy, "start", mock.MagicMock())
    def test_new_connection_close(self, mock_create_connection, mock_socket):
        fake_client = mock.MagicMock()
        fake_server = mock.MagicMock()
        mock_socket.side_effect = [fake_client]
        mock_create_connection.side_effect = [fake_server]
        client = proxy.socket()
        self.assertIs(client, fake_client)
        self.assertIsNone(self.proxy.new_connection(client, b""))

    @mock.patch.object(proxy, "socket")
    @mock.patch.object(proxy, "create_connection")
    @mock.patch.object(proxy.ResponseProxy, "stop", mock.MagicMock())
    @mock.patch.object(proxy.ResponseProxy, "start", mock.MagicMock())
    def test_new_connection_OSError(self, mock_create_connection, mock_socket):
        fake_client = mock.MagicMock()
        fake_server = mock.MagicMock()
        mock_socket.side_effect = [fake_client]
        mock_create_connection.side_effect = [fake_server]
        client = proxy.socket()
        self.assertIs(client, fake_client)
        fake_client.recv.side_effect = [b"test", OSError()]
        self.assertIsNone(self.proxy.new_connection(client, b"test"))

    @mock.patch.object(proxy, "socket")
    @mock.patch.object(proxy, "create_connection")
    @mock.patch.object(proxy.ResponseProxy, "stop", mock.MagicMock())
    @mock.patch.object(proxy.ResponseProxy, "start", mock.MagicMock())
    def test_new_connection_Exception(self, mock_create_connection, mock_socket):  # noqa:501
        fake_client = mock.MagicMock()
        fake_server = mock.MagicMock()
        mock_socket.side_effect = [fake_client]
        mock_create_connection.side_effect = [fake_server]
        client = proxy.socket()
        self.assertIs(client, fake_client)
        fake_client.recv.side_effect = [b"test", proxy.timeout(), Exception()]
        self.assertIsNone(self.proxy.new_connection(client, b"test"))


if __name__ == "__main__":
    unittest.main()



# coding:utf-8

__project__ = "xser"
__version__ = "0.5.1"
__urlhome__ = "https://github.com/bondbox/xser/"
__description__ = "Socket and HTTP Server"

# author
__author__ = "Mingzhe Zou"
__author_email__ = "zoumingzhe@outlook.com"



# coding=utf-8

from os.path import dirname
from os.path import join
from urllib.parse import urljoin

from setuptools import find_packages
from setuptools import setup
from setuptools.command.install import install

from xserver.attribute import __author__
from xserver.attribute import __author_email__
from xserver.attribute import __description__
from xserver.attribute import __project__
from xserver.attribute import __urlhome__
from xserver.attribute import __version__

__urlcode__ = __urlhome__
__urldocs__ = __urlhome__
__urlbugs__ = urljoin(__urlhome__, "issues")


def all_requirements():
    def read_requirements(path: str):
        with open(path, "r", encoding="utf-8") as rhdl:
            return rhdl.read().splitlines()

    path: str = join(dirname(__file__), "requirements.txt")
    requirements = read_requirements(path)
    return requirements


class CustomInstallCommand(install):
    """Customized setuptools install command"""

    def run(self):
        install.run(self)  # Run the standard installation
        # Execute your custom code after installation


setup(
    name=__project__,
    version=__version__,
    description=__description__,
    url=__urlhome__,
    author=__author__,
    author_email=__author_email__,
    project_urls={"Source Code": __urlcode__,
                  "Bug Tracker": __urlbugs__,
                  "Documentation": __urldocs__},
    packages=find_packages(include=["xserver*"], exclude=["xserver.unittest"]),
    install_requires=all_requirements(),
    cmdclass={
        "install": CustomInstallCommand,
    }
)
